// Copyright 2004 Adam Megacz, see the COPYING file for licensing [GPL]
package org.ibex;

import java.io.*;
import org.ibex.js.*;
import org.ibex.util.*;
import org.bouncycastle.util.encoders.Base64;

/** Singleton class that provides all functionality in the ibex.* namespace */
public final class Ibex extends JS.Cloneable {

    // FIXME remove this
    private final JS rr;

    public Ibex(Stream rr) { this.rr = bless(rr); }

    public JS resolveString(String str, boolean permitAbsolute) throws JSExn {
        if (str.indexOf("://") != -1) {
            if (permitAbsolute) return (Stream)url2res(str);
            throw new JSExn("absolute URL " + str + " not permitted here");
        }
        // root-relative
        //JS ret = (JS)getAndTriggerTraps("");
        //FIXME
        JS ret = rr;
        while(str.indexOf('.') != -1) {
            String path = str.substring(0, str.indexOf('.'));
            str = str.substring(str.indexOf('.') + 1);
            ret = (JS)ret.get(path);
        }
        if (!"".equals(str)) ret = (JS)ret.get(str);
        return ret;
    }

    /** lets us put multi-level get/put/call keys all in the same method */
    private class Sub extends JS {
        String key;
        Sub(String key) { this.key = key; }
        public void put(Object key, Object val) throws JSExn { Ibex.this.put(this.key + "." + key, val); }
        public Object get(Object key) throws JSExn { return Ibex.this.get(this.key + "." + key); }
        public Object call(Object a0, Object a1, Object a2, Object[] rest, int nargs) throws JSExn {
            return Ibex.this.callMethod(this.key, a0, a1, a2, rest, nargs);
        }
        public Object callMethod(Object method, Object a0, Object a1, Object a2, Object[] rest, int nargs) throws JSExn {
            return Ibex.this.callMethod(this.key + "." + method, a0, a1, a2, rest, nargs);
        }
    }
    private Cache subCache = new Cache(20);
    private Sub getSub(String s) {
        Sub ret = (Sub)subCache.get(s);
        if (ret == null) subCache.put(s, ret = new Sub(s));
        return ret;
    }

    public Object get(Object name) throws JSExn {
        if (name instanceof String && ((String)name).length() == 0) return rr;
        //#switch(name)
        case "math": return ibexMath;
        case "string": return ibexString;
        case "date": return METHOD;
        case "box": return new Box();
        case "clone": return METHOD;
        case "bless": return METHOD;
        case "regexp": return METHOD;
        case "ui": return getSub("ui");
        case "ui.font": return getSub("ui.font");
        case "ui.font.wait": return METHOD;
        case "ui.font.width": return METHOD;
        case "ui.font.height": return METHOD;
        case "ui.font.sansserif": return Main.builtin.get("fonts/vera/Vera.ttf");
        case "ui.font.monospace": return Main.builtin.get("fonts/vera/VeraMono.ttf");
        case "ui.font.serif": return Main.builtin.get("fonts/vera/VeraSe.ttf");
        case "ui.browser": return METHOD;
        case "ui.mouse": return getSub("ui.mouse");
        case "ui.mouse.button":
            if (Surface.button1 && !Surface.button2 && !Surface.button3) return N(1);
            else if (!Surface.button1 && Surface.button2 && !Surface.button3) return N(2);
            else if (!Surface.button1 && !Surface.button2 && Surface.button3) return N(3);
            else return ZERO;
        case "ui.key": return getSub("ui.key");
        case "ui.key.name": return getSub("ui.key.name");
        case "ui.key.name.alt": return Platform.altKeyName();
        case "ui.key.alt": return Surface.alt ? T : F;
        case "ui.key.control": return Surface.control ? T : F;
        case "ui.key.shift": return Surface.shift ? T : F;
        case "ui.clipboard": return Platform.getClipBoard();
        case "ui.maxdim": return N(Short.MAX_VALUE);
        case "ui.screen": return getSub("ui.screen");
        case "ui.screen.width": return N(Platform.getScreenWidth());
        case "ui.screen.height": return N(Platform.getScreenHeight());
        case "undocumented": return getSub("undocumented");
        case "undocumented.initialOrigin": return Main.origin;
        case "undocumented.initialTemplate": return Main.initialTemplate;
        case "thread": return getSub("thread");
        case "thread.yield": return METHOD;
        case "thread.sleep": return METHOD;
        case "stream": return getSub("stream");
        case "stream.homedir": return url2res("file:" + System.getProperty("user.home"));
        case "stream.tempdir": return url2res("file:" + System.getProperty("java.io.tempdir"));
        case "stream.watch": return METHOD;
        case "stream.unzip": return METHOD;
        case "stream.uncab": return METHOD;
        case "stream.cache": return METHOD;
        case "stream.url": return METHOD;
        case "stream.parse.html": return METHOD;
        case "stream.parse.xml": return METHOD;
        case "stream.parse.utf8": return METHOD;
        case "net": return getSub("net");
        case "net.rpc": return getSub("net.rpc");
        case "net.rpc.xml": return METHOD;
        case "net.rpc.soap": return METHOD;
        case "log": return getSub("log");
        case "log.debug": return METHOD;
        case "log.info": return METHOD;
        case "log.warn": return METHOD;
        case "log.error": return METHOD;
        case "crypto": return getSub("crypto");
        case "crypto.rsa": return METHOD;
        case "crypto.md5": return METHOD;
        case "crypto.sha1": return METHOD;
        case "crypto.rc4": return METHOD;
        //#end
        return super.get(name);
    }

    public void put(Object name, final Object value) throws JSExn {
        //#switch(name)
        case "thread": Scheduler.add((Scheduler.Task)value); return;
        case "ui.clipboard": Platform.setClipBoard((String)value); return;
        case "ui.frame": Platform.createSurface((Box)value, true, true); return;
        case "ui.window": Platform.createSurface((Box)value, false, true); return;
        case "undocumented.proxyAuthorization":
            HTTP.Proxy.Authorization.authorization = value.toString();
            HTTP.Proxy.Authorization.waitingForUser.release();
            return;
        //#end
        throw new JSExn("attempted to put unknown property: ibex."+name);
    }

    public Object callMethod(Object name, Object a, Object b, Object c, Object[] rest, int nargs) throws JSExn {
        try {
            //#switch(name)
            case "date": return new JSDate(a, b, c, rest, nargs);
            case "net.rpc.soap": return new SOAP((String)a, "", (String)b, (String)c);
                // FIXME support object dumping
            case "log.debug":    JS.debug(a== null ? "**null**" : a.toString()); return null;
            case "log.info":     JS.info(a== null ? "**null**" : a.toString()); return null;
            case "log.warn":     JS.warn(a== null ? "**null**" : a.toString()); return null;
            case "log.error":    JS.error(a== null ? "**null**" : a.toString()); return null;
            //#end
 
            switch (nargs) {
                case 0:
                    //#switch(name)
                    case "thread.yield": sleep(0); return null;
                    //#end
                    break;
                case 1:
                    //#switch(name)
                    case "clone":
                        if (!(a instanceof JS.Cloneable)) throw new JSExn("cannot clone a " + a.getClass().getName());
                        return ((JS.Cloneable)a).jsclone();
                    case "bless": return bless((JS)a);
                    case "ui.browser": Platform.newBrowserWindow((String)a); return null;
                    case "stream.unzip": return new Stream.Zip((Stream)a);
                    case "stream.uncab": return new Stream.Cab((Stream)a);
                    case "stream.cache":
                        try { return new Stream.CachedStream((Stream)a, "resources", true); }
                        catch (Stream.NotCacheableException e) { throw new JSExn("this resource cannot be cached"); }
                    case "stream.url": {
                        String url = (String)a;
                        if (url.startsWith("http://")) return new Stream.HTTP(url);
                        else if (url.startsWith("https://")) return new Stream.HTTP(url);
                        else if (url.startsWith("data:")) return new Stream.ByteArray(Base64.decode(url.substring(5)), null);
                        else if (url.startsWith("utf8:")) return new Stream.ByteArray(url.substring(5).getBytes(), null);
                        else if (url.startsWith("file:")) {
                            // FIXME
                            Platform.fileDialog(url.substring(5), false);
                        }
                        throw new JSExn("invalid resource specifier " + url);
                    }
                    case "thread.sleep": sleep(JS.toInt(a)); return null;
                    case "regexp": return new JSRegexp(a, null);
                    case "net.rpc.xml": return new XMLRPC((String)a, "");
                    case "crypto.rsa": /* FEATURE */ return null;
                    case "crypto.md5": /* FEATURE */ return null;
                    case "crypto.sha1": /* FEATURE */ return null;
                    case "crypto.rc4": /* FEATURE */ return null;
                    case "stream.parse.html": throw new JSExn("not implemented yet"); //return null;
                    case "stream.parse.xml": new XMLHelper((JS)b).doParse((JS)a); return null;
                        // FIXME backgrounding
                    case "stream.parse.utf8": try { return new String(InputStreamToByteArray.convert(Stream.getInputStream(a))); }
                                              catch (Exception e) { Log.warn(this, e); }
                    //#end
                    break;
                case 2:
                    //#switch(name)
                    case "stream.watch": return new Stream.ProgressWatcher((Stream)a, (JS)b);
                    case "regexp": return new JSRegexp(a, b);
                    //#end
                case 3:
                    //#switch(name)
                    case "ui.font.height": return N(Font.getFont((Stream)a, JS.toInt(b)).textheight((String)c));
                    case "ui.font.wait":
                        try {
                            JS.UnpauseCallback callback = JS.pause();
                            Font font = Font.getFont((Stream)a, JS.toInt(b));
                            if (font.rasterizeGlyphs((String)c, null, 0,0,0,0,0,0,0, callback) != -1) Scheduler.add(callback);
                        } catch (JS.NotPauseableException npe) {
                            JS.error("can't wait for a font in a foreground thread");
                        }
                        return null;
                    case "ui.font.width": return N(Font.getFont((Stream)a, JS.toInt(b)).textwidth((String)c));
                    //#end
                    break;
            }
        } catch (RuntimeException e) {
            // FIXME: maybe JSExn should take a second argument, Exception
            Log.warn(this, "ibex."+name+"() threw: " + e);
            throw new JSExn("invalid argument for ibex object method "+name+"()");
        }

        throw new JSExn("invalid number of arguments ("+nargs+") for ibex object method "+name+"()");
    }

    public Stream url2res(String url) throws JSExn {
        if (url.startsWith("http://")) return new Stream.HTTP(url);
        else if (url.startsWith("https://")) return new Stream.HTTP(url);
        else if (url.startsWith("data:")) return new Stream.ByteArray(Base64.decode(url.substring(5)), null);
        else if (url.startsWith("utf8:")) return new Stream.ByteArray(url.substring(5).getBytes(), null);
        else throw new JSExn("invalid resource specifier " + url);
        // FIXME support file:// via dialog boxes
    }

    public static void sleep(final int i) throws JSExn {
        try {
            final JS.UnpauseCallback callback = JS.pause();
            // FEATURE use a single sleeper thread
            new Thread() { public void run() {
                try { Thread.sleep(i); } catch (InterruptedException e) { }
                Scheduler.add(callback);
            } }.start();
        } catch (JS.NotPauseableException npe) {
            throw new JSExn("you cannot sleep or yield in the foreground thread");
        }
    }
    
    public static final JSMath ibexMath = new JSMath() {
            private JS gs = new JSScope.Global();
            public Object get(Object key) throws JSExn {
                //#switch(key)
                case "isNaN": return gs.get("isNaN");
                case "isFinite": return gs.get("isFinite");
                case "NaN": return gs.get("NaN");
                case "Infinity": return gs.get("Infinity");
                //#end
                return super.get(key);
            }
        };

    public static final JS ibexString = new JS() {
            private JS gs = new JSScope.Global();
            public void put(Object key, Object val) { }
            public Object get(Object key) throws JSExn {
                //#switch(key)
                case "parseInt": return gs.get("parseInt");
                case "parseFloat": return gs.get("parseFloat");
                case "decodeURI": return gs.get("decodeURI");
                case "decodeURIComponent": return gs.get("decodeURIComponent");
                case "encodeURI": return gs.get("encodeURI");
                case "encodeURIComponent": return gs.get("encodeURIComponent");
                case "escape": return gs.get("escape");
                case "unescape": return gs.get("unescape");
                case "fromCharCode": return gs.get("stringFromCharCode");
                //#end
                return null;
            }
        };

    private class XMLHelper extends XML {
        private class Wrapper extends XML.Exn { public JSExn wrapee; public Wrapper(JSExn jse) { super(""); wrapee = jse; } }
        private JS characters, whitespace, endElement, startElement;
        public XMLHelper(JS b) throws JSExn {
            super(BUFFER_SIZE);
            startElement = (JS)b.getAndTriggerTraps("startElement");
            endElement   = (JS)b.getAndTriggerTraps("endElement");
            characters   = (JS)b.getAndTriggerTraps("characters");
            whitespace   = (JS)b.getAndTriggerTraps("whitespace");
        }

        public void startElement(XML.Element c) throws XML.Exn { try {
                JS attrs = new JS();
                // FIXME attribute URIs? add an additional hash?
                for(int i=0; i<c.getAttrLen(); i++) attrs.put(c.getAttrKey(i), c.getAttrVal(i));
                startElement.call(c.getLocalName(), attrs, c.getUri(), null, 3);
        } catch (JSExn jse) { throw new Wrapper(jse); } }

        public void endElement(XML.Element c) throws XML.Exn { try {
                endElement.call(c.getLocalName(), c.getUri(), null, null, 2);
        } catch (JSExn jse) { throw new Wrapper(jse); } }

        public void characters(char[] ch, int start, int length) throws XML.Exn { try {
                characters.call(new String(ch, start, length), null, null, null, 1);
        } catch (JSExn jse) { throw new Wrapper(jse); } }

        public void whitespace(char[] ch, int start, int length) throws XML.Exn { try {
                whitespace.call(new String(ch, start, length), null, null, null, 1);
        } catch (JSExn jse) { throw new Wrapper(jse); } }

        public void doParse(JS s) throws JSExn {
            try { 
                parse(new BufferedReader(new InputStreamReader(Stream.getInputStream(s))));
            } catch (Wrapper e) {
                throw e.wrapee;
            } catch (XML.Exn e) {
                throw new JSExn("error parsing XML: " + e.toString());
            } catch (IOException e) {
                if (Log.on) Log.info(this, "IO Exception while reading from file");
                if (Log.on) Log.info(this, e);
                throw new JSExn("error reading from Resource");
            }
        }
    }

    // FEATURE: move this into builtin.xwar
    public Blessing bless(JS b) { return new Ibex.Blessing((JS.Cloneable)b, this, null, null); }
    public static class Blessing extends JS.Clone {
        private Ibex ibex;
        private Template t = null;
        public Object parentkey = null;
        public Blessing parent = null;
        private Hash cache = new Hash();
        public Blessing(JS.Cloneable clonee, Ibex ibex, Blessing parent, Object parentkey) {
            super(clonee); this.ibex = ibex; this.parentkey = parentkey; this.parent = parent; }
        public Object get(Object key) throws JSExn {
            if (key.equals("")) return ((Object)getStatic());
            if (cache.get(key) != null) return cache.get(key);
            Object ret = new Blessing((JS.Cloneable)clonee.get(key), ibex, this, key);
            cache.put(key, ret);
            return ret;
        }
        public static Blessing getBlessing(Object o) {
            if (!(o instanceof JS)) return null;
            JS js = (JS)o;
            while (js instanceof JS.Clone && !(js instanceof Blessing)) js = ((JS.Clone)js).getClonee();
            if (!(js instanceof Blessing)) return null;
            return (Blessing)js;
        }
        public InputStream getImage() throws JSExn {
            try {
                InputStream in = Stream.getInputStream(this);
                if (in != null) return in;
            } catch (IOException e) { /* DELIBERATE */ }
            String[] exts = new String[] { ".png", ".jpeg", ".gif" };
            for (int i=0; i < exts.length; i++)
                try {
                    InputStream in = Stream.getInputStream(parent.get(parentkey + exts[i]));
                    if (in != null) return in;
                } catch (IOException f) { /* DELIBERATE */ }
            return null;
        }
        public JSScope getStatic() {
            try {
                if (t == null) t = Template.buildTemplate(parentkey + ".t", parent.get(parentkey + ".t"), ibex);
                return t.staticScope;
            } catch (Exception e) {
                Log.error(this, e);
                return null;
            }
        }
        public Object call(Object a, Object b, Object c, Object[] rest, int nargs) throws JSExn {
            // GROSS hack
            if (nargs != 1 && nargs != 9999) throw new JSExn("FIXME can only call with one arg");
            getStatic();
            if (t == null) throw new JSExn("No such template " + parentkey);
            if (nargs == 9999) return t;
            t.apply((Box)a);
            return a;
        }
    }

}
